# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .config_setting_contract_bandwidth_v0 import ConfigSettingContractBandwidthV0
from .config_setting_contract_compute_v0 import ConfigSettingContractComputeV0
from .config_setting_contract_events_v0 import ConfigSettingContractEventsV0
from .config_setting_contract_execution_lanes_v0 import (
    ConfigSettingContractExecutionLanesV0,
)
from .config_setting_contract_historical_data_v0 import (
    ConfigSettingContractHistoricalDataV0,
)
from .config_setting_contract_ledger_cost_v0 import ConfigSettingContractLedgerCostV0
from .config_setting_id import ConfigSettingID
from .contract_cost_params import ContractCostParams
from .eviction_iterator import EvictionIterator
from .state_archival_settings import StateArchivalSettings
from .uint32 import Uint32
from .uint64 import Uint64

__all__ = ["ConfigSettingEntry"]


class ConfigSettingEntry:
    """
    XDR Source Code::

        union ConfigSettingEntry switch (ConfigSettingID configSettingID)
        {
        case CONFIG_SETTING_CONTRACT_MAX_SIZE_BYTES:
            uint32 contractMaxSizeBytes;
        case CONFIG_SETTING_CONTRACT_COMPUTE_V0:
            ConfigSettingContractComputeV0 contractCompute;
        case CONFIG_SETTING_CONTRACT_LEDGER_COST_V0:
            ConfigSettingContractLedgerCostV0 contractLedgerCost;
        case CONFIG_SETTING_CONTRACT_HISTORICAL_DATA_V0:
            ConfigSettingContractHistoricalDataV0 contractHistoricalData;
        case CONFIG_SETTING_CONTRACT_EVENTS_V0:
            ConfigSettingContractEventsV0 contractEvents;
        case CONFIG_SETTING_CONTRACT_BANDWIDTH_V0:
            ConfigSettingContractBandwidthV0 contractBandwidth;
        case CONFIG_SETTING_CONTRACT_COST_PARAMS_CPU_INSTRUCTIONS:
            ContractCostParams contractCostParamsCpuInsns;
        case CONFIG_SETTING_CONTRACT_COST_PARAMS_MEMORY_BYTES:
            ContractCostParams contractCostParamsMemBytes;
        case CONFIG_SETTING_CONTRACT_DATA_KEY_SIZE_BYTES:
            uint32 contractDataKeySizeBytes;
        case CONFIG_SETTING_CONTRACT_DATA_ENTRY_SIZE_BYTES:
            uint32 contractDataEntrySizeBytes;
        case CONFIG_SETTING_STATE_ARCHIVAL:
            StateArchivalSettings stateArchivalSettings;
        case CONFIG_SETTING_CONTRACT_EXECUTION_LANES:
            ConfigSettingContractExecutionLanesV0 contractExecutionLanes;
        case CONFIG_SETTING_BUCKETLIST_SIZE_WINDOW:
            uint64 bucketListSizeWindow<>;
        case CONFIG_SETTING_EVICTION_ITERATOR:
            EvictionIterator evictionIterator;
        };
    """

    def __init__(
        self,
        config_setting_id: ConfigSettingID,
        contract_max_size_bytes: Uint32 = None,
        contract_compute: ConfigSettingContractComputeV0 = None,
        contract_ledger_cost: ConfigSettingContractLedgerCostV0 = None,
        contract_historical_data: ConfigSettingContractHistoricalDataV0 = None,
        contract_events: ConfigSettingContractEventsV0 = None,
        contract_bandwidth: ConfigSettingContractBandwidthV0 = None,
        contract_cost_params_cpu_insns: ContractCostParams = None,
        contract_cost_params_mem_bytes: ContractCostParams = None,
        contract_data_key_size_bytes: Uint32 = None,
        contract_data_entry_size_bytes: Uint32 = None,
        state_archival_settings: StateArchivalSettings = None,
        contract_execution_lanes: ConfigSettingContractExecutionLanesV0 = None,
        bucket_list_size_window: List[Uint64] = None,
        eviction_iterator: EvictionIterator = None,
    ) -> None:
        _expect_max_length = 4294967295
        if (
            bucket_list_size_window
            and len(bucket_list_size_window) > _expect_max_length
        ):
            raise ValueError(
                f"The maximum length of `bucket_list_size_window` should be {_expect_max_length}, but got {len(bucket_list_size_window)}."
            )
        self.config_setting_id = config_setting_id
        self.contract_max_size_bytes = contract_max_size_bytes
        self.contract_compute = contract_compute
        self.contract_ledger_cost = contract_ledger_cost
        self.contract_historical_data = contract_historical_data
        self.contract_events = contract_events
        self.contract_bandwidth = contract_bandwidth
        self.contract_cost_params_cpu_insns = contract_cost_params_cpu_insns
        self.contract_cost_params_mem_bytes = contract_cost_params_mem_bytes
        self.contract_data_key_size_bytes = contract_data_key_size_bytes
        self.contract_data_entry_size_bytes = contract_data_entry_size_bytes
        self.state_archival_settings = state_archival_settings
        self.contract_execution_lanes = contract_execution_lanes
        self.bucket_list_size_window = bucket_list_size_window
        self.eviction_iterator = eviction_iterator

    def pack(self, packer: Packer) -> None:
        self.config_setting_id.pack(packer)
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_MAX_SIZE_BYTES
        ):
            if self.contract_max_size_bytes is None:
                raise ValueError("contract_max_size_bytes should not be None.")
            self.contract_max_size_bytes.pack(packer)
            return
        if self.config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_COMPUTE_V0:
            if self.contract_compute is None:
                raise ValueError("contract_compute should not be None.")
            self.contract_compute.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_LEDGER_COST_V0
        ):
            if self.contract_ledger_cost is None:
                raise ValueError("contract_ledger_cost should not be None.")
            self.contract_ledger_cost.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_HISTORICAL_DATA_V0
        ):
            if self.contract_historical_data is None:
                raise ValueError("contract_historical_data should not be None.")
            self.contract_historical_data.pack(packer)
            return
        if self.config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_EVENTS_V0:
            if self.contract_events is None:
                raise ValueError("contract_events should not be None.")
            self.contract_events.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_BANDWIDTH_V0
        ):
            if self.contract_bandwidth is None:
                raise ValueError("contract_bandwidth should not be None.")
            self.contract_bandwidth.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_COST_PARAMS_CPU_INSTRUCTIONS
        ):
            if self.contract_cost_params_cpu_insns is None:
                raise ValueError("contract_cost_params_cpu_insns should not be None.")
            self.contract_cost_params_cpu_insns.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_COST_PARAMS_MEMORY_BYTES
        ):
            if self.contract_cost_params_mem_bytes is None:
                raise ValueError("contract_cost_params_mem_bytes should not be None.")
            self.contract_cost_params_mem_bytes.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_DATA_KEY_SIZE_BYTES
        ):
            if self.contract_data_key_size_bytes is None:
                raise ValueError("contract_data_key_size_bytes should not be None.")
            self.contract_data_key_size_bytes.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_DATA_ENTRY_SIZE_BYTES
        ):
            if self.contract_data_entry_size_bytes is None:
                raise ValueError("contract_data_entry_size_bytes should not be None.")
            self.contract_data_entry_size_bytes.pack(packer)
            return
        if self.config_setting_id == ConfigSettingID.CONFIG_SETTING_STATE_ARCHIVAL:
            if self.state_archival_settings is None:
                raise ValueError("state_archival_settings should not be None.")
            self.state_archival_settings.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_EXECUTION_LANES
        ):
            if self.contract_execution_lanes is None:
                raise ValueError("contract_execution_lanes should not be None.")
            self.contract_execution_lanes.pack(packer)
            return
        if (
            self.config_setting_id
            == ConfigSettingID.CONFIG_SETTING_BUCKETLIST_SIZE_WINDOW
        ):
            if self.bucket_list_size_window is None:
                raise ValueError("bucket_list_size_window should not be None.")
            packer.pack_uint(len(self.bucket_list_size_window))
            for bucket_list_size_window_item in self.bucket_list_size_window:
                bucket_list_size_window_item.pack(packer)
            return
        if self.config_setting_id == ConfigSettingID.CONFIG_SETTING_EVICTION_ITERATOR:
            if self.eviction_iterator is None:
                raise ValueError("eviction_iterator should not be None.")
            self.eviction_iterator.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingEntry:
        config_setting_id = ConfigSettingID.unpack(unpacker)
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_MAX_SIZE_BYTES:
            contract_max_size_bytes = Uint32.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_max_size_bytes=contract_max_size_bytes,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_COMPUTE_V0:
            contract_compute = ConfigSettingContractComputeV0.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id, contract_compute=contract_compute
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_LEDGER_COST_V0:
            contract_ledger_cost = ConfigSettingContractLedgerCostV0.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_ledger_cost=contract_ledger_cost,
            )
        if (
            config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_HISTORICAL_DATA_V0
        ):
            contract_historical_data = ConfigSettingContractHistoricalDataV0.unpack(
                unpacker
            )
            return cls(
                config_setting_id=config_setting_id,
                contract_historical_data=contract_historical_data,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_EVENTS_V0:
            contract_events = ConfigSettingContractEventsV0.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id, contract_events=contract_events
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_BANDWIDTH_V0:
            contract_bandwidth = ConfigSettingContractBandwidthV0.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_bandwidth=contract_bandwidth,
            )
        if (
            config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_COST_PARAMS_CPU_INSTRUCTIONS
        ):
            contract_cost_params_cpu_insns = ContractCostParams.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_cost_params_cpu_insns=contract_cost_params_cpu_insns,
            )
        if (
            config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_COST_PARAMS_MEMORY_BYTES
        ):
            contract_cost_params_mem_bytes = ContractCostParams.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_cost_params_mem_bytes=contract_cost_params_mem_bytes,
            )
        if (
            config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_DATA_KEY_SIZE_BYTES
        ):
            contract_data_key_size_bytes = Uint32.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_data_key_size_bytes=contract_data_key_size_bytes,
            )
        if (
            config_setting_id
            == ConfigSettingID.CONFIG_SETTING_CONTRACT_DATA_ENTRY_SIZE_BYTES
        ):
            contract_data_entry_size_bytes = Uint32.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                contract_data_entry_size_bytes=contract_data_entry_size_bytes,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_STATE_ARCHIVAL:
            state_archival_settings = StateArchivalSettings.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id,
                state_archival_settings=state_archival_settings,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_CONTRACT_EXECUTION_LANES:
            contract_execution_lanes = ConfigSettingContractExecutionLanesV0.unpack(
                unpacker
            )
            return cls(
                config_setting_id=config_setting_id,
                contract_execution_lanes=contract_execution_lanes,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_BUCKETLIST_SIZE_WINDOW:
            length = unpacker.unpack_uint()
            bucket_list_size_window = []
            for _ in range(length):
                bucket_list_size_window.append(Uint64.unpack(unpacker))
            return cls(
                config_setting_id=config_setting_id,
                bucket_list_size_window=bucket_list_size_window,
            )
        if config_setting_id == ConfigSettingID.CONFIG_SETTING_EVICTION_ITERATOR:
            eviction_iterator = EvictionIterator.unpack(unpacker)
            return cls(
                config_setting_id=config_setting_id, eviction_iterator=eviction_iterator
            )
        return cls(config_setting_id=config_setting_id)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.config_setting_id,
                self.contract_max_size_bytes,
                self.contract_compute,
                self.contract_ledger_cost,
                self.contract_historical_data,
                self.contract_events,
                self.contract_bandwidth,
                self.contract_cost_params_cpu_insns,
                self.contract_cost_params_mem_bytes,
                self.contract_data_key_size_bytes,
                self.contract_data_entry_size_bytes,
                self.state_archival_settings,
                self.contract_execution_lanes,
                self.bucket_list_size_window,
                self.eviction_iterator,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.config_setting_id == other.config_setting_id
            and self.contract_max_size_bytes == other.contract_max_size_bytes
            and self.contract_compute == other.contract_compute
            and self.contract_ledger_cost == other.contract_ledger_cost
            and self.contract_historical_data == other.contract_historical_data
            and self.contract_events == other.contract_events
            and self.contract_bandwidth == other.contract_bandwidth
            and self.contract_cost_params_cpu_insns
            == other.contract_cost_params_cpu_insns
            and self.contract_cost_params_mem_bytes
            == other.contract_cost_params_mem_bytes
            and self.contract_data_key_size_bytes == other.contract_data_key_size_bytes
            and self.contract_data_entry_size_bytes
            == other.contract_data_entry_size_bytes
            and self.state_archival_settings == other.state_archival_settings
            and self.contract_execution_lanes == other.contract_execution_lanes
            and self.bucket_list_size_window == other.bucket_list_size_window
            and self.eviction_iterator == other.eviction_iterator
        )

    def __repr__(self):
        out = []
        out.append(f"config_setting_id={self.config_setting_id}")
        (
            out.append(f"contract_max_size_bytes={self.contract_max_size_bytes}")
            if self.contract_max_size_bytes is not None
            else None
        )
        (
            out.append(f"contract_compute={self.contract_compute}")
            if self.contract_compute is not None
            else None
        )
        (
            out.append(f"contract_ledger_cost={self.contract_ledger_cost}")
            if self.contract_ledger_cost is not None
            else None
        )
        (
            out.append(f"contract_historical_data={self.contract_historical_data}")
            if self.contract_historical_data is not None
            else None
        )
        (
            out.append(f"contract_events={self.contract_events}")
            if self.contract_events is not None
            else None
        )
        (
            out.append(f"contract_bandwidth={self.contract_bandwidth}")
            if self.contract_bandwidth is not None
            else None
        )
        (
            out.append(
                f"contract_cost_params_cpu_insns={self.contract_cost_params_cpu_insns}"
            )
            if self.contract_cost_params_cpu_insns is not None
            else None
        )
        (
            out.append(
                f"contract_cost_params_mem_bytes={self.contract_cost_params_mem_bytes}"
            )
            if self.contract_cost_params_mem_bytes is not None
            else None
        )
        (
            out.append(
                f"contract_data_key_size_bytes={self.contract_data_key_size_bytes}"
            )
            if self.contract_data_key_size_bytes is not None
            else None
        )
        (
            out.append(
                f"contract_data_entry_size_bytes={self.contract_data_entry_size_bytes}"
            )
            if self.contract_data_entry_size_bytes is not None
            else None
        )
        (
            out.append(f"state_archival_settings={self.state_archival_settings}")
            if self.state_archival_settings is not None
            else None
        )
        (
            out.append(f"contract_execution_lanes={self.contract_execution_lanes}")
            if self.contract_execution_lanes is not None
            else None
        )
        (
            out.append(f"bucket_list_size_window={self.bucket_list_size_window}")
            if self.bucket_list_size_window is not None
            else None
        )
        (
            out.append(f"eviction_iterator={self.eviction_iterator}")
            if self.eviction_iterator is not None
            else None
        )
        return f"<ConfigSettingEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .uint32 import Uint32

__all__ = ["ConfigSettingContractComputeV0"]


class ConfigSettingContractComputeV0:
    """
    XDR Source Code::

        struct ConfigSettingContractComputeV0
        {
            // Maximum instructions per ledger
            int64 ledgerMaxInstructions;
            // Maximum instructions per transaction
            int64 txMaxInstructions;
            // Cost of 10000 instructions
            int64 feeRatePerInstructionsIncrement;

            // Memory limit per transaction. Unlike instructions, there is no fee
            // for memory, just the limit.
            uint32 txMemoryLimit;
        };
    """

    def __init__(
        self,
        ledger_max_instructions: Int64,
        tx_max_instructions: Int64,
        fee_rate_per_instructions_increment: Int64,
        tx_memory_limit: Uint32,
    ) -> None:
        self.ledger_max_instructions = ledger_max_instructions
        self.tx_max_instructions = tx_max_instructions
        self.fee_rate_per_instructions_increment = fee_rate_per_instructions_increment
        self.tx_memory_limit = tx_memory_limit

    def pack(self, packer: Packer) -> None:
        self.ledger_max_instructions.pack(packer)
        self.tx_max_instructions.pack(packer)
        self.fee_rate_per_instructions_increment.pack(packer)
        self.tx_memory_limit.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingContractComputeV0:
        ledger_max_instructions = Int64.unpack(unpacker)
        tx_max_instructions = Int64.unpack(unpacker)
        fee_rate_per_instructions_increment = Int64.unpack(unpacker)
        tx_memory_limit = Uint32.unpack(unpacker)
        return cls(
            ledger_max_instructions=ledger_max_instructions,
            tx_max_instructions=tx_max_instructions,
            fee_rate_per_instructions_increment=fee_rate_per_instructions_increment,
            tx_memory_limit=tx_memory_limit,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingContractComputeV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingContractComputeV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_max_instructions,
                self.tx_max_instructions,
                self.fee_rate_per_instructions_increment,
                self.tx_memory_limit,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_max_instructions == other.ledger_max_instructions
            and self.tx_max_instructions == other.tx_max_instructions
            and self.fee_rate_per_instructions_increment
            == other.fee_rate_per_instructions_increment
            and self.tx_memory_limit == other.tx_memory_limit
        )

    def __repr__(self):
        out = [
            f"ledger_max_instructions={self.ledger_max_instructions}",
            f"tx_max_instructions={self.tx_max_instructions}",
            f"fee_rate_per_instructions_increment={self.fee_rate_per_instructions_increment}",
            f"tx_memory_limit={self.tx_memory_limit}",
        ]
        return f"<ConfigSettingContractComputeV0 [{', '.join(out)}]>"



from typing import TYPE_CHECKING, Union

from ..exceptions import SdkError

if TYPE_CHECKING:
    from .assembled_transaction import AssembledTransaction
    from .assembled_transaction_async import AssembledTransactionAsync

__all__ = [
    "AssembledTransactionError",
    "RestorationFailureError",
    "NotYetSimulatedError",
    "SimulationFailedError",
    "ExpiredStateError",
    "NoSignatureNeededError",
    "NeedsMoreSignaturesError",
    "SendTransactionFailedError",
    "TransactionFailedError",
    "TransactionStillPendingError",
]


class AssembledTransactionError(SdkError):
    """Raised when an assembled transaction fails."""

    def __init__(
        self,
        message,
        assembled_transaction: Union[
            "AssembledTransaction", "AssembledTransactionAsync"
        ],
    ) -> None:
        super().__init__(message)
        self.assembled_transaction = assembled_transaction


class RestorationFailureError(AssembledTransactionError):
    """Raised when a restoration fails."""


class NotYetSimulatedError(AssembledTransactionError):
    """Raised when trying to get the result of a transaction that has not been simulated yet."""


class SimulationFailedError(AssembledTransactionError):
    """Raised when a simulation fails."""


class ExpiredStateError(AssembledTransactionError):
    """Raised when the state has expired."""


class NoSignatureNeededError(AssembledTransactionError):
    """Raised when no signature is needed."""


class NeedsMoreSignaturesError(AssembledTransactionError):
    """Raised when more signatures are needed."""


class SendTransactionFailedError(AssembledTransactionError):
    """Raised when invoking `send_transaction` fails."""


class TransactionFailedError(AssembledTransactionError):
    """Raised when invoking `get_transaction` fails."""


class TransactionStillPendingError(AssembledTransactionError):
    """Raised when the transaction is still pending."""



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .hash import Hash
from .sc_address_type import SCAddressType

__all__ = ["SCAddress"]


class SCAddress:
    """
    XDR Source Code::

        union SCAddress switch (SCAddressType type)
        {
        case SC_ADDRESS_TYPE_ACCOUNT:
            AccountID accountId;
        case SC_ADDRESS_TYPE_CONTRACT:
            Hash contractId;
        };
    """

    def __init__(
        self,
        type: SCAddressType,
        account_id: AccountID = None,
        contract_id: Hash = None,
    ) -> None:
        self.type = type
        self.account_id = account_id
        self.contract_id = contract_id

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SCAddressType.SC_ADDRESS_TYPE_ACCOUNT:
            if self.account_id is None:
                raise ValueError("account_id should not be None.")
            self.account_id.pack(packer)
            return
        if self.type == SCAddressType.SC_ADDRESS_TYPE_CONTRACT:
            if self.contract_id is None:
                raise ValueError("contract_id should not be None.")
            self.contract_id.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCAddress:
        type = SCAddressType.unpack(unpacker)
        if type == SCAddressType.SC_ADDRESS_TYPE_ACCOUNT:
            account_id = AccountID.unpack(unpacker)
            return cls(type=type, account_id=account_id)
        if type == SCAddressType.SC_ADDRESS_TYPE_CONTRACT:
            contract_id = Hash.unpack(unpacker)
            return cls(type=type, contract_id=contract_id)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCAddress:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCAddress:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.account_id,
                self.contract_id,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.account_id == other.account_id
            and self.contract_id == other.contract_id
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"account_id={self.account_id}")
            if self.account_id is not None
            else None
        )
        (
            out.append(f"contract_id={self.contract_id}")
            if self.contract_id is not None
            else None
        )
        return f"<SCAddress [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ErrorCode"]


class ErrorCode(IntEnum):
    """
    XDR Source Code::

        enum ErrorCode
        {
            ERR_MISC = 0, // Unspecific error
            ERR_DATA = 1, // Malformed data
            ERR_CONF = 2, // Misconfiguration error
            ERR_AUTH = 3, // Authentication failure
            ERR_LOAD = 4  // System overloaded
        };
    """

    ERR_MISC = 0
    ERR_DATA = 1
    ERR_CONF = 2
    ERR_AUTH = 3
    ERR_LOAD = 4

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ErrorCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ErrorCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ErrorCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseOrderbookCallBuilder"]


class BaseOrderbookCallBuilder(BaseCallBuilder):
    """Creates a new :class:`OrderbookCallBuilder` pointed to server defined by horizon_url.

    See `Orderbook <https://developers.stellar.org/api/aggregations/order-books/>`__ for more information.

    :param selling: Asset being sold
    :param buying: Asset being bought
    :param horizon_url: Horizon server URL.
    """

    def __init__(self, selling: Asset, buying: Asset, **kwargs) -> None:
        super().__init__(**kwargs)
        self.endpoint: str = "order_book"
        params = {
            "selling_asset_type": selling.type,
            "selling_asset_code": None if selling.is_native() else selling.code,
            "selling_asset_issuer": selling.issuer,
            "buying_asset_type": buying.type,
            "buying_asset_code": None if buying.is_native() else buying.code,
            "buying_asset_issuer": buying.issuer,
        }
        self._add_query_params(params)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .uint32 import Uint32
from .value import Value

__all__ = ["SCPBallot"]


class SCPBallot:
    """
    XDR Source Code::

        struct SCPBallot
        {
            uint32 counter; // n
            Value value;    // x
        };
    """

    def __init__(
        self,
        counter: Uint32,
        value: Value,
    ) -> None:
        self.counter = counter
        self.value = value

    def pack(self, packer: Packer) -> None:
        self.counter.pack(packer)
        self.value.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCPBallot:
        counter = Uint32.unpack(unpacker)
        value = Value.unpack(unpacker)
        return cls(
            counter=counter,
            value=value,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCPBallot:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCPBallot:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.counter,
                self.value,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.counter == other.counter and self.value == other.value

    def __repr__(self):
        out = [
            f"counter={self.counter}",
            f"value={self.value}",
        ]
        return f"<SCPBallot [{', '.join(out)}]>"



from decimal import Decimal
from typing import Any, Dict, List, Union

from .account import Account
from .asset import Asset
from .base_server import BaseServer
from .call_builder.call_builder_async import *
from .client.aiohttp_client import AiohttpClient
from .client.base_async_client import BaseAsyncClient
from .client.response import Response
from .exceptions import NotFoundError, raise_request_exception
from .fee_bump_transaction import FeeBumpTransaction
from .fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from .keypair import Keypair
from .memo import NoneMemo
from .muxed_account import MuxedAccount
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope
from .utils import MUXED_ACCOUNT_STARTING_LETTER, urljoin_with_query

__all__ = ["ServerAsync"]


class ServerAsync(BaseServer):
    """ServerAsync handles the network connection to a `Horizon <https://developers.stellar.org/api/introduction/>`_
    instance and exposes an interface for requests to that instance.

    An example::

        import asyncio
        from stellar_sdk import ServerAsync

        async def example():
            async with ServerAsync("https://horizon-testnet.stellar.org") as server:
                resp = await server.transactions().limit(10).order(desc=True).call()
                print(resp)

        asyncio.run(example())

    :param horizon_url: Horizon Server URL
        (ex. ``"https://horizon-testnet.stellar.org"`` for test network,
        ``"https://horizon.stellar.org"`` for public network)
    :param client: Http client used to send the request
    """

    def __init__(
        self,
        horizon_url: str = "https://horizon-testnet.stellar.org/",
        client: BaseAsyncClient = None,
    ) -> None:
        self.horizon_url: str = horizon_url

        if not client:
            client = AiohttpClient()
        self._client: BaseAsyncClient = client

    async def submit_transaction(
        self,
        transaction_envelope: Union[
            TransactionEnvelope, FeeBumpTransactionEnvelope, str
        ],
        skip_memo_required_check: bool = False,
    ) -> Dict[str, Any]:
        """Submits a transaction to the network.

        :param transaction_envelope: :class:`stellar_sdk.transaction_envelope.TransactionEnvelope` object
            or base64 encoded xdr
        :param skip_memo_required_check: Allow skipping memo
        :return: the response from horizon
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
            :exc:`AccountRequiresMemoError <stellar_sdk.sep.exceptions.AccountRequiresMemoError>`
        """
        url = urljoin_with_query(self.horizon_url, "transactions")
        xdr, tx = self._get_xdr_and_transaction_from_transaction_envelope(
            transaction_envelope
        )
        if not skip_memo_required_check:
            await self._check_memo_required(tx)
        data = {"tx": xdr}
        resp = await self._client.post(url=url, data=data)
        assert isinstance(resp, Response)
        raise_request_exception(resp)
        return resp.json()

    async def submit_transaction_async(
        self,
        transaction_envelope: Union[
            TransactionEnvelope, FeeBumpTransactionEnvelope, str
        ],
        skip_memo_required_check: bool = False,
    ) -> Dict[str, Any]:
        """Submits an asynchronous transaction to the network. Unlike the synchronous version, which blocks
        and waits for the transaction to be ingested in Horizon, this endpoint relays the response from
        core directly back to the user.

        See `Horizon Documentation - Submit a Transaction Asynchronously <https://developers.stellar.org/docs/data/horizon/api-reference/resources/submit-async-transaction>`_

        :param transaction_envelope: :class:`stellar_sdk.transaction_envelope.TransactionEnvelope` object
            or base64 encoded xdr
        :param skip_memo_required_check: Allow skipping memo
        :return: the response from horizon
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
            :exc:`AccountRequiresMemoError <stellar_sdk.sep.exceptions.AccountRequiresMemoError>`
        """
        url = urljoin_with_query(self.horizon_url, "transactions_async")
        xdr, tx = self._get_xdr_and_transaction_from_transaction_envelope(
            transaction_envelope
        )
        if not skip_memo_required_check:
            await self._check_memo_required(tx)
        data = {"tx": xdr}
        resp = await self._client.post(url=url, data=data)
        assert isinstance(resp, Response)
        raise_request_exception(resp)
        return resp.json()

    async def load_account(
        self, account_id: Union[MuxedAccount, Keypair, str]
    ) -> Account:
        """Fetches an account's most current base state (like sequence) in the ledger and then creates
        and returns an :class:`stellar_sdk.account.Account` object.

        If you want to get complete account information, please
        use :func:`stellar_sdk.server.Server.accounts`.

        :param account_id: The account to load.
        :return: an :class:`stellar_sdk.account.Account` object.
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
        """
        if isinstance(account_id, str):
            account_id = MuxedAccount.from_account(account_id)
        elif isinstance(account_id, Keypair):
            account_id = MuxedAccount.from_account(account_id.public_key)
        else:
            account_id = account_id
        resp = await self.accounts().account_id(account_id=account_id.account_id).call()
        sequence = int(resp["sequence"])
        account = Account(account=account_id, sequence=sequence, raw_data=resp)
        return account

    async def _check_memo_required(
        self, transaction: Union[Transaction, FeeBumpTransaction]
    ) -> None:
        if isinstance(transaction, FeeBumpTransaction):
            transaction = transaction.inner_transaction_envelope.transaction
        if not (transaction.memo is None or isinstance(transaction.memo, NoneMemo)):
            return
        for index, destination in self._get_check_memo_required_destinations(
            transaction
        ):
            if destination.startswith(MUXED_ACCOUNT_STARTING_LETTER):
                continue
            try:
                account_resp = await self.accounts().account_id(destination).call()
            except NotFoundError:
                continue
            self._check_destination_memo(account_resp, index, destination)

    async def fetch_base_fee(self) -> int:
        """Fetch the base fee. Since this hits the server, if the server call fails,
        you might get an error. You should be prepared to use a default value if that happens.

        :return: the base fee
        :raises:
            :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
            :exc:`NotFoundError <stellar_sdk.exceptions.NotFoundError>`
            :exc:`BadRequestError <stellar_sdk.exceptions.BadRequestError>`
            :exc:`BadResponseError <stellar_sdk.exceptions.BadResponseError>`
            :exc:`UnknownRequestError <stellar_sdk.exceptions.UnknownRequestError>`
        """
        latest_ledger = await self.ledgers().order(desc=True).limit(1).call()
        base_fee = self._handle_base_fee(latest_ledger)
        return base_fee

    def root(self) -> RootCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.RootCallBuilder` object configured
            by a current Horizon server configuration.
        """
        return RootCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def accounts(self) -> AccountsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.AccountsCallBuilder` object configured
            by a current Horizon server configuration.
        """
        return AccountsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def assets(self) -> AssetsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.AssetsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return AssetsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def claimable_balances(self) -> ClaimableBalancesCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.ClaimableBalancesCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return ClaimableBalancesCallBuilder(
            horizon_url=self.horizon_url, client=self._client
        )

    def data(self, account_id: str, data_name: str) -> DataCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.DataCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return DataCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            account_id=account_id,
            data_name=data_name,
        )

    def effects(self) -> EffectsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.EffectsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return EffectsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def fee_stats(self) -> FeeStatsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.FeeStatsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return FeeStatsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def ledgers(self) -> LedgersCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.LedgersCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return LedgersCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def liquidity_pools(self) -> LiquidityPoolsBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.LiquidityPoolsBuilder` object configured by
            a current Horizon server configuration.
        """
        return LiquidityPoolsBuilder(horizon_url=self.horizon_url, client=self._client)

    def offers(self) -> OffersCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.OffersCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OffersCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def operations(self) -> OperationsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.OperationsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OperationsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def orderbook(self, selling: Asset, buying: Asset) -> OrderbookCallBuilder:
        """
        :param selling: Asset being sold
        :param buying: Asset being bought
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.OrderbookCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return OrderbookCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            buying=buying,
            selling=selling,
        )

    def strict_receive_paths(
        self,
        source: Union[str, List[Asset]],
        destination_asset: Asset,
        destination_amount: Union[str, Decimal],
    ) -> StrictReceivePathsCallBuilder:
        """
        :param source: The sender's account ID or a list of Assets. Any returned path must use a source that the sender can hold.
        :param destination_asset: The destination asset.
        :param destination_amount: The amount, denominated in the destination asset, that any returned path should be able to satisfy.
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.StrictReceivePathsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return StrictReceivePathsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            source=source,
            destination_asset=destination_asset,
            destination_amount=destination_amount,
        )

    def strict_send_paths(
        self,
        source_asset: Asset,
        source_amount: Union[str, Decimal],
        destination: Union[str, List[Asset]],
    ) -> StrictSendPathsCallBuilder:
        """
        :param source_asset: The asset to be sent.
        :param source_amount: The amount, denominated in the source asset, that any returned path should be able to satisfy.
        :param destination: The destination account or the destination assets.
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.StrictReceivePathsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return StrictSendPathsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            source_asset=source_asset,
            source_amount=source_amount,
            destination=destination,
        )

    def payments(self) -> PaymentsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.PaymentsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return PaymentsCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def trade_aggregations(
        self,
        base: Asset,
        counter: Asset,
        resolution: int,
        start_time: int = None,
        end_time: int = None,
        offset: int = None,
    ) -> TradeAggregationsCallBuilder:
        """
        :param base: base asset
        :param counter: counter asset
        :param resolution: segment duration as millis since epoch. *Supported values
            are 1 minute (60000), 5 minutes (300000), 15 minutes (900000),
            1 hour (3600000), 1 day (86400000) and 1 week (604800000).*
        :param start_time: lower time boundary represented as millis since epoch
        :param end_time: upper time boundary represented as millis since epoch
        :param offset: segments can be offset using this parameter.
            Expressed in milliseconds. *Can only be used if the resolution is greater than 1 hour.
            Value must be in whole hours, less than the provided resolution, and less than 24 hours.*
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.TradeAggregationsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TradeAggregationsCallBuilder(
            horizon_url=self.horizon_url,
            client=self._client,
            base=base,
            counter=counter,
            start_time=start_time,
            end_time=end_time,
            resolution=resolution,
            offset=offset,
        )

    def trades(self) -> TradesCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.TradesCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TradesCallBuilder(horizon_url=self.horizon_url, client=self._client)

    def transactions(self) -> TransactionsCallBuilder:
        """
        :return: New :class:`stellar_sdk.call_builder.call_builder_async.TransactionsCallBuilder` object configured by
            a current Horizon server configuration.
        """
        return TransactionsCallBuilder(
            horizon_url=self.horizon_url, client=self._client
        )

    async def close(self) -> None:
        """Close underlying connector, and release all acquired resources."""
        await self._client.close()

    async def __aenter__(self) -> "ServerAsync":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    def __repr__(self):
        return f"<ServerAsync [horizon_url={self.horizon_url}, client={self._client}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_error_code import SCErrorCode
from .sc_error_type import SCErrorType
from .uint32 import Uint32

__all__ = ["SCError"]


class SCError:
    """
    XDR Source Code::

        union SCError switch (SCErrorType type)
        {
        case SCE_CONTRACT:
            uint32 contractCode;
        case SCE_WASM_VM:
        case SCE_CONTEXT:
        case SCE_STORAGE:
        case SCE_OBJECT:
        case SCE_CRYPTO:
        case SCE_EVENTS:
        case SCE_BUDGET:
        case SCE_VALUE:
        case SCE_AUTH:
            SCErrorCode code;
        };
    """

    def __init__(
        self,
        type: SCErrorType,
        contract_code: Uint32 = None,
        code: SCErrorCode = None,
    ) -> None:
        self.type = type
        self.contract_code = contract_code
        self.code = code

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SCErrorType.SCE_CONTRACT:
            if self.contract_code is None:
                raise ValueError("contract_code should not be None.")
            self.contract_code.pack(packer)
            return
        if self.type == SCErrorType.SCE_WASM_VM:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_CONTEXT:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_STORAGE:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_OBJECT:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_CRYPTO:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_EVENTS:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_BUDGET:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_VALUE:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return
        if self.type == SCErrorType.SCE_AUTH:
            if self.code is None:
                raise ValueError("code should not be None.")
            self.code.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCError:
        type = SCErrorType.unpack(unpacker)
        if type == SCErrorType.SCE_CONTRACT:
            contract_code = Uint32.unpack(unpacker)
            return cls(type=type, contract_code=contract_code)
        if type == SCErrorType.SCE_WASM_VM:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_CONTEXT:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_STORAGE:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_OBJECT:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_CRYPTO:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_EVENTS:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_BUDGET:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_VALUE:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        if type == SCErrorType.SCE_AUTH:
            code = SCErrorCode.unpack(unpacker)
            return cls(type=type, code=code)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCError:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCError:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.contract_code,
                self.code,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.contract_code == other.contract_code
            and self.code == other.code
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"contract_code={self.contract_code}")
            if self.contract_code is not None
            else None
        )
        out.append(f"code={self.code}") if self.code is not None else None
        return f"<SCError [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .cold_archive_bucket_entry import ColdArchiveBucketEntry
from .ledger_key import LedgerKey
from .proof_level import ProofLevel

__all__ = ["ExistenceProofBody"]


class ExistenceProofBody:
    """
    XDR Source Code::

        struct ExistenceProofBody
        {
            LedgerKey keysToProve<>;

            // Bounds for each key being proved, where bound[n]
            // corresponds to keysToProve[n]
            ColdArchiveBucketEntry lowBoundEntries<>;
            ColdArchiveBucketEntry highBoundEntries<>;

            // Vector of vectors, where proofLevels[level]
            // contains all HashNodes that correspond with that level
            ProofLevel proofLevels<>;
        };
    """

    def __init__(
        self,
        keys_to_prove: List[LedgerKey],
        low_bound_entries: List[ColdArchiveBucketEntry],
        high_bound_entries: List[ColdArchiveBucketEntry],
        proof_levels: List[ProofLevel],
    ) -> None:
        _expect_max_length = 4294967295
        if keys_to_prove and len(keys_to_prove) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `keys_to_prove` should be {_expect_max_length}, but got {len(keys_to_prove)}."
            )
        _expect_max_length = 4294967295
        if low_bound_entries and len(low_bound_entries) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `low_bound_entries` should be {_expect_max_length}, but got {len(low_bound_entries)}."
            )
        _expect_max_length = 4294967295
        if high_bound_entries and len(high_bound_entries) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `high_bound_entries` should be {_expect_max_length}, but got {len(high_bound_entries)}."
            )
        _expect_max_length = 4294967295
        if proof_levels and len(proof_levels) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `proof_levels` should be {_expect_max_length}, but got {len(proof_levels)}."
            )
        self.keys_to_prove = keys_to_prove
        self.low_bound_entries = low_bound_entries
        self.high_bound_entries = high_bound_entries
        self.proof_levels = proof_levels

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.keys_to_prove))
        for keys_to_prove_item in self.keys_to_prove:
            keys_to_prove_item.pack(packer)
        packer.pack_uint(len(self.low_bound_entries))
        for low_bound_entries_item in self.low_bound_entries:
            low_bound_entries_item.pack(packer)
        packer.pack_uint(len(self.high_bound_entries))
        for high_bound_entries_item in self.high_bound_entries:
            high_bound_entries_item.pack(packer)
        packer.pack_uint(len(self.proof_levels))
        for proof_levels_item in self.proof_levels:
            proof_levels_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ExistenceProofBody:
        length = unpacker.unpack_uint()
        keys_to_prove = []
        for _ in range(length):
            keys_to_prove.append(LedgerKey.unpack(unpacker))
        length = unpacker.unpack_uint()
        low_bound_entries = []
        for _ in range(length):
            low_bound_entries.append(ColdArchiveBucketEntry.unpack(unpacker))
        length = unpacker.unpack_uint()
        high_bound_entries = []
        for _ in range(length):
            high_bound_entries.append(ColdArchiveBucketEntry.unpack(unpacker))
        length = unpacker.unpack_uint()
        proof_levels = []
        for _ in range(length):
            proof_levels.append(ProofLevel.unpack(unpacker))
        return cls(
            keys_to_prove=keys_to_prove,
            low_bound_entries=low_bound_entries,
            high_bound_entries=high_bound_entries,
            proof_levels=proof_levels,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ExistenceProofBody:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ExistenceProofBody:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.keys_to_prove,
                self.low_bound_entries,
                self.high_bound_entries,
                self.proof_levels,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.keys_to_prove == other.keys_to_prove
            and self.low_bound_entries == other.low_bound_entries
            and self.high_bound_entries == other.high_bound_entries
            and self.proof_levels == other.proof_levels
        )

    def __repr__(self):
        out = [
            f"keys_to_prove={self.keys_to_prove}",
            f"low_bound_entries={self.low_bound_entries}",
            f"high_bound_entries={self.high_bound_entries}",
            f"proof_levels={self.proof_levels}",
        ]
        return f"<ExistenceProofBody [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .signer_key import SignerKey
from .uint32 import Uint32

__all__ = ["Signer"]


class Signer:
    """
    XDR Source Code::

        struct Signer
        {
            SignerKey key;
            uint32 weight; // really only need 1 byte
        };
    """

    def __init__(
        self,
        key: SignerKey,
        weight: Uint32,
    ) -> None:
        self.key = key
        self.weight = weight

    def pack(self, packer: Packer) -> None:
        self.key.pack(packer)
        self.weight.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Signer:
        key = SignerKey.unpack(unpacker)
        weight = Uint32.unpack(unpacker)
        return cls(
            key=key,
            weight=weight,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Signer:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Signer:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.key,
                self.weight,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.key == other.key and self.weight == other.weight

    def __repr__(self):
        out = [
            f"key={self.key}",
            f"weight={self.weight}",
        ]
        return f"<Signer [{', '.join(out)}]>"



import base64
import binascii
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from ..utils import raise_if_not_valid_balance_id
from .operation import Operation

__all__ = ["ClawbackClaimableBalance"]


class ClawbackClaimableBalance(Operation):
    """The :class:`ClawbackClaimableBalance` object, which represents a ClawbackClaimableBalance operation on
    Stellar's network.

    Claws back a claimable balance

    Threshold: Medium

    See `Clawback Claimable Balance <https://developers.stellar.org/docs/start/list-of-operations/#clawback-claimable-balance>`_ for more information.

    :param balance_id: The claimable balance ID to be clawed back.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.CLAWBACK_CLAIMABLE_BALANCE
    )

    def __init__(
        self, balance_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> None:
        super().__init__(source)
        self.balance_id: str = balance_id
        raise_if_not_valid_balance_id(self.balance_id, "balance_id")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        balance_id_bytes: bytes = binascii.unhexlify(self.balance_id)
        balance_id = stellar_xdr.ClaimableBalanceID.from_xdr_bytes(balance_id_bytes)
        clawback_claimable_balance_op = stellar_xdr.ClawbackClaimableBalanceOp(
            balance_id=balance_id
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            clawback_claimable_balance_op=clawback_claimable_balance_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "ClawbackClaimableBalance":
        """Creates a :class:`ClawbackClaimableBalance` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.clawback_claimable_balance_op is not None
        balance_id_bytes = base64.b64decode(
            xdr_object.body.clawback_claimable_balance_op.to_xdr()
        )
        balance_id = binascii.hexlify(balance_id_bytes).decode()
        op = cls(balance_id=balance_id, source=source)
        return op

    def __repr__(self):
        return f"<ClawbackClaimableBalance [balance_id={self.balance_id}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .sequence_number import SequenceNumber
from .uint32 import Uint32

__all__ = ["HashIDPreimageOperationID"]


class HashIDPreimageOperationID:
    """
    XDR Source Code::

        struct
            {
                AccountID sourceAccount;
                SequenceNumber seqNum;
                uint32 opNum;
            }
    """

    def __init__(
        self,
        source_account: AccountID,
        seq_num: SequenceNumber,
        op_num: Uint32,
    ) -> None:
        self.source_account = source_account
        self.seq_num = seq_num
        self.op_num = op_num

    def pack(self, packer: Packer) -> None:
        self.source_account.pack(packer)
        self.seq_num.pack(packer)
        self.op_num.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HashIDPreimageOperationID:
        source_account = AccountID.unpack(unpacker)
        seq_num = SequenceNumber.unpack(unpacker)
        op_num = Uint32.unpack(unpacker)
        return cls(
            source_account=source_account,
            seq_num=seq_num,
            op_num=op_num,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HashIDPreimageOperationID:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HashIDPreimageOperationID:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.source_account,
                self.seq_num,
                self.op_num,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.source_account == other.source_account
            and self.seq_num == other.seq_num
            and self.op_num == other.op_num
        )

    def __repr__(self):
        out = [
            f"source_account={self.source_account}",
            f"seq_num={self.seq_num}",
            f"op_num={self.op_num}",
        ]
        return f"<HashIDPreimageOperationID [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .archival_proof_body import ArchivalProofBody
from .uint32 import Uint32

__all__ = ["ArchivalProof"]


class ArchivalProof:
    """
    XDR Source Code::

        struct ArchivalProof
        {
            uint32 epoch; // AST Subtree for this proof

            union switch (ArchivalProofType t)
            {
            case EXISTENCE:
                NonexistenceProofBody nonexistenceProof;
            case NONEXISTENCE:
                ExistenceProofBody existenceProof;
            } body;
        };
    """

    def __init__(
        self,
        epoch: Uint32,
        body: ArchivalProofBody,
    ) -> None:
        self.epoch = epoch
        self.body = body

    def pack(self, packer: Packer) -> None:
        self.epoch.pack(packer)
        self.body.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ArchivalProof:
        epoch = Uint32.unpack(unpacker)
        body = ArchivalProofBody.unpack(unpacker)
        return cls(
            epoch=epoch,
            body=body,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ArchivalProof:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ArchivalProof:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.epoch,
                self.body,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.epoch == other.epoch and self.body == other.body

    def __repr__(self):
        out = [
            f"epoch={self.epoch}",
            f"body={self.body}",
        ]
        return f"<ArchivalProof [{', '.join(out)}]>"



from ...call_builder.base import BaseDataCallBuilder
from ...call_builder.call_builder_async.base_call_builder import BaseCallBuilder
from ...client.base_async_client import BaseAsyncClient

__all__ = ["DataCallBuilder"]


class DataCallBuilder(BaseCallBuilder, BaseDataCallBuilder):
    """Creates a new :class:`DataCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.ServerAsync.data`.

    See `Retrieve an Account's Data <https://developers.stellar.org/api/resources/accounts/data/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param account_id: account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
    :param data_name: Key name
    """

    def __init__(
        self,
        horizon_url: str,
        client: BaseAsyncClient,
        account_id: str,
        data_name: str,
    ) -> None:
        super().__init__(
            horizon_url=horizon_url,
            client=client,
            account_id=account_id,
            data_name=data_name,
        )



from typing import Union

from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseEffectsCallBuilder"]


class BaseEffectsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`EffectsCallBuilder` pointed to server defined by horizon_url.

    See `List All Effects <https://developers.stellar.org/api/resources/effects/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "effects"

    def for_account(self, account_id: str):
        """This endpoint represents all effects that changed a given account. It will return relevant
        effects from the creation of the account to the current ledger.

        See `Retrieve an Account's Effects <https://developers.stellar.org/api/resources/accounts/effects/>`__ for more information.

        :param account_id: account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: this EffectCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/effects"
        return self

    def for_ledger(self, sequence: Union[int, str]):
        """Effects are the specific ways that the ledger was changed by any operation.
        This endpoint represents all effects that occurred in the given ledger.

        See `Retrieve a Ledger's Effects <https://developers.stellar.org/api/resources/ledgers/effects/>`__ for more information.

        :param sequence: ledger sequence
        :return: this EffectCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}/effects"
        return self

    def for_transaction(self, transaction_hash: str):
        """This endpoint represents all effects that occurred as a result of a given transaction.

        See `Retrieve a Transaction's Effects <https://developers.stellar.org/api/resources/transactions/effects/>`__ for more information.

        :param transaction_hash: transaction hash
        :return: this EffectCallBuilder instance
        """
        self.endpoint = f"transactions/{transaction_hash}/effects"
        return self

    def for_operation(self, operation_id: Union[int, str]):
        """This endpoint represents all effects that occurred as a result of a given operation.

        See `Retrieve an Operation's Effects <https://developers.stellar.org/api/resources/operations/effects/>`__ for more information.

        :param operation_id: operation ID
        :return: this EffectCallBuilder instance
        """
        self.endpoint = f"operations/{operation_id}/effects"
        return self

    def for_liquidity_pool(self, liquidity_pool_id: str):
        """This endpoint represents all effects that occurred as a result of a given liquidity pool.

        See `Liquidity Pools - Retrieve related Effects <https://developers.stellar.org/api/resources/liquiditypools/effects/>`__ for more information.

        :param liquidity_pool_id: The ID of the liquidity pool in hex string.
        :return: this EffectsCallBuilder instance
        """
        self.endpoint = f"liquidity_pools/{liquidity_pool_id}/effects"
        return self



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .ledger_header_history_entry import LedgerHeaderHistoryEntry
from .scp_history_entry import SCPHistoryEntry
from .transaction_result_meta import TransactionResultMeta
from .transaction_set import TransactionSet
from .upgrade_entry_meta import UpgradeEntryMeta

__all__ = ["LedgerCloseMetaV0"]


class LedgerCloseMetaV0:
    """
    XDR Source Code::

        struct LedgerCloseMetaV0
        {
            LedgerHeaderHistoryEntry ledgerHeader;
            // NB: txSet is sorted in "Hash order"
            TransactionSet txSet;

            // NB: transactions are sorted in apply order here
            // fees for all transactions are processed first
            // followed by applying transactions
            TransactionResultMeta txProcessing<>;

            // upgrades are applied last
            UpgradeEntryMeta upgradesProcessing<>;

            // other misc information attached to the ledger close
            SCPHistoryEntry scpInfo<>;
        };
    """

    def __init__(
        self,
        ledger_header: LedgerHeaderHistoryEntry,
        tx_set: TransactionSet,
        tx_processing: List[TransactionResultMeta],
        upgrades_processing: List[UpgradeEntryMeta],
        scp_info: List[SCPHistoryEntry],
    ) -> None:
        _expect_max_length = 4294967295
        if tx_processing and len(tx_processing) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `tx_processing` should be {_expect_max_length}, but got {len(tx_processing)}."
            )
        _expect_max_length = 4294967295
        if upgrades_processing and len(upgrades_processing) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `upgrades_processing` should be {_expect_max_length}, but got {len(upgrades_processing)}."
            )
        _expect_max_length = 4294967295
        if scp_info and len(scp_info) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `scp_info` should be {_expect_max_length}, but got {len(scp_info)}."
            )
        self.ledger_header = ledger_header
        self.tx_set = tx_set
        self.tx_processing = tx_processing
        self.upgrades_processing = upgrades_processing
        self.scp_info = scp_info

    def pack(self, packer: Packer) -> None:
        self.ledger_header.pack(packer)
        self.tx_set.pack(packer)
        packer.pack_uint(len(self.tx_processing))
        for tx_processing_item in self.tx_processing:
            tx_processing_item.pack(packer)
        packer.pack_uint(len(self.upgrades_processing))
        for upgrades_processing_item in self.upgrades_processing:
            upgrades_processing_item.pack(packer)
        packer.pack_uint(len(self.scp_info))
        for scp_info_item in self.scp_info:
            scp_info_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerCloseMetaV0:
        ledger_header = LedgerHeaderHistoryEntry.unpack(unpacker)
        tx_set = TransactionSet.unpack(unpacker)
        length = unpacker.unpack_uint()
        tx_processing = []
        for _ in range(length):
            tx_processing.append(TransactionResultMeta.unpack(unpacker))
        length = unpacker.unpack_uint()
        upgrades_processing = []
        for _ in range(length):
            upgrades_processing.append(UpgradeEntryMeta.unpack(unpacker))
        length = unpacker.unpack_uint()
        scp_info = []
        for _ in range(length):
            scp_info.append(SCPHistoryEntry.unpack(unpacker))
        return cls(
            ledger_header=ledger_header,
            tx_set=tx_set,
            tx_processing=tx_processing,
            upgrades_processing=upgrades_processing,
            scp_info=scp_info,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerCloseMetaV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerCloseMetaV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_header,
                self.tx_set,
                self.tx_processing,
                self.upgrades_processing,
                self.scp_info,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_header == other.ledger_header
            and self.tx_set == other.tx_set
            and self.tx_processing == other.tx_processing
            and self.upgrades_processing == other.upgrades_processing
            and self.scp_info == other.scp_info
        )

    def __repr__(self):
        out = [
            f"ledger_header={self.ledger_header}",
            f"tx_set={self.tx_set}",
            f"tx_processing={self.tx_processing}",
            f"upgrades_processing={self.upgrades_processing}",
            f"scp_info={self.scp_info}",
        ]
        return f"<LedgerCloseMetaV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["PublicKeyType"]


class PublicKeyType(IntEnum):
    """
    XDR Source Code::

        enum PublicKeyType
        {
            PUBLIC_KEY_TYPE_ED25519 = KEY_TYPE_ED25519
        };
    """

    PUBLIC_KEY_TYPE_ED25519 = 0

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PublicKeyType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PublicKeyType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PublicKeyType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Generator

from .response import Response

__all__ = ["BaseSyncClient"]


class BaseSyncClient(metaclass=ABCMeta):
    """This is an abstract class,
    and if you want to implement your own synchronous client, you **must** implement this class.

    """

    @abstractmethod
    def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """

    @abstractmethod
    def post(
        self,
        url: str,
        data: Dict[str, str] = None,
        json_data: Dict[str, Any] = None,
    ) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :param json_data: the json data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """

    @abstractmethod
    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://developers.stellar.org/api/introduction/response-format/>`__

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`_

        :param url: the request url
        :param params: the request params
        :return: a dict Generator for server response
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """

    @abstractmethod
    def close(self):
        pass



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .bucket_metadata import BucketMetadata
from .hot_archive_bucket_entry_type import HotArchiveBucketEntryType
from .ledger_entry import LedgerEntry
from .ledger_key import LedgerKey

__all__ = ["HotArchiveBucketEntry"]


class HotArchiveBucketEntry:
    """
    XDR Source Code::

        union HotArchiveBucketEntry switch (HotArchiveBucketEntryType type)
        {
        case HOT_ARCHIVE_ARCHIVED:
            LedgerEntry archivedEntry;

        case HOT_ARCHIVE_LIVE:
        case HOT_ARCHIVE_DELETED:
            LedgerKey key;
        case HOT_ARCHIVE_METAENTRY:
            BucketMetadata metaEntry;
        };
    """

    def __init__(
        self,
        type: HotArchiveBucketEntryType,
        archived_entry: LedgerEntry = None,
        key: LedgerKey = None,
        meta_entry: BucketMetadata = None,
    ) -> None:
        self.type = type
        self.archived_entry = archived_entry
        self.key = key
        self.meta_entry = meta_entry

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == HotArchiveBucketEntryType.HOT_ARCHIVE_ARCHIVED:
            if self.archived_entry is None:
                raise ValueError("archived_entry should not be None.")
            self.archived_entry.pack(packer)
            return
        if self.type == HotArchiveBucketEntryType.HOT_ARCHIVE_LIVE:
            if self.key is None:
                raise ValueError("key should not be None.")
            self.key.pack(packer)
            return
        if self.type == HotArchiveBucketEntryType.HOT_ARCHIVE_DELETED:
            if self.key is None:
                raise ValueError("key should not be None.")
            self.key.pack(packer)
            return
        if self.type == HotArchiveBucketEntryType.HOT_ARCHIVE_METAENTRY:
            if self.meta_entry is None:
                raise ValueError("meta_entry should not be None.")
            self.meta_entry.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HotArchiveBucketEntry:
        type = HotArchiveBucketEntryType.unpack(unpacker)
        if type == HotArchiveBucketEntryType.HOT_ARCHIVE_ARCHIVED:
            archived_entry = LedgerEntry.unpack(unpacker)
            return cls(type=type, archived_entry=archived_entry)
        if type == HotArchiveBucketEntryType.HOT_ARCHIVE_LIVE:
            key = LedgerKey.unpack(unpacker)
            return cls(type=type, key=key)
        if type == HotArchiveBucketEntryType.HOT_ARCHIVE_DELETED:
            key = LedgerKey.unpack(unpacker)
            return cls(type=type, key=key)
        if type == HotArchiveBucketEntryType.HOT_ARCHIVE_METAENTRY:
            meta_entry = BucketMetadata.unpack(unpacker)
            return cls(type=type, meta_entry=meta_entry)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HotArchiveBucketEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HotArchiveBucketEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.archived_entry,
                self.key,
                self.meta_entry,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.archived_entry == other.archived_entry
            and self.key == other.key
            and self.meta_entry == other.meta_entry
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"archived_entry={self.archived_entry}")
            if self.archived_entry is not None
            else None
        )
        out.append(f"key={self.key}") if self.key is not None else None
        (
            out.append(f"meta_entry={self.meta_entry}")
            if self.meta_entry is not None
            else None
        )
        return f"<HotArchiveBucketEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .soroban_address_credentials import SorobanAddressCredentials
from .soroban_credentials_type import SorobanCredentialsType

__all__ = ["SorobanCredentials"]


class SorobanCredentials:
    """
    XDR Source Code::

        union SorobanCredentials switch (SorobanCredentialsType type)
        {
        case SOROBAN_CREDENTIALS_SOURCE_ACCOUNT:
            void;
        case SOROBAN_CREDENTIALS_ADDRESS:
            SorobanAddressCredentials address;
        };
    """

    def __init__(
        self,
        type: SorobanCredentialsType,
        address: SorobanAddressCredentials = None,
    ) -> None:
        self.type = type
        self.address = address

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SorobanCredentialsType.SOROBAN_CREDENTIALS_SOURCE_ACCOUNT:
            return
        if self.type == SorobanCredentialsType.SOROBAN_CREDENTIALS_ADDRESS:
            if self.address is None:
                raise ValueError("address should not be None.")
            self.address.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SorobanCredentials:
        type = SorobanCredentialsType.unpack(unpacker)
        if type == SorobanCredentialsType.SOROBAN_CREDENTIALS_SOURCE_ACCOUNT:
            return cls(type=type)
        if type == SorobanCredentialsType.SOROBAN_CREDENTIALS_ADDRESS:
            address = SorobanAddressCredentials.unpack(unpacker)
            return cls(type=type, address=address)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SorobanCredentials:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SorobanCredentials:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.address,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.type == other.type and self.address == other.address

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"address={self.address}") if self.address is not None else None
        return f"<SorobanCredentials [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Boolean
from .uint32 import Uint32
from .uint64 import Uint64

__all__ = ["EvictionIterator"]


class EvictionIterator:
    """
    XDR Source Code::

        struct EvictionIterator {
            uint32 bucketListLevel;
            bool isCurrBucket;
            uint64 bucketFileOffset;
        };
    """

    def __init__(
        self,
        bucket_list_level: Uint32,
        is_curr_bucket: bool,
        bucket_file_offset: Uint64,
    ) -> None:
        self.bucket_list_level = bucket_list_level
        self.is_curr_bucket = is_curr_bucket
        self.bucket_file_offset = bucket_file_offset

    def pack(self, packer: Packer) -> None:
        self.bucket_list_level.pack(packer)
        Boolean(self.is_curr_bucket).pack(packer)
        self.bucket_file_offset.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> EvictionIterator:
        bucket_list_level = Uint32.unpack(unpacker)
        is_curr_bucket = Boolean.unpack(unpacker)
        bucket_file_offset = Uint64.unpack(unpacker)
        return cls(
            bucket_list_level=bucket_list_level,
            is_curr_bucket=is_curr_bucket,
            bucket_file_offset=bucket_file_offset,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> EvictionIterator:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> EvictionIterator:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.bucket_list_level,
                self.is_curr_bucket,
                self.bucket_file_offset,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.bucket_list_level == other.bucket_list_level
            and self.is_curr_bucket == other.is_curr_bucket
            and self.bucket_file_offset == other.bucket_file_offset
        )

    def __repr__(self):
        out = [
            f"bucket_list_level={self.bucket_list_level}",
            f"is_curr_bucket={self.is_curr_bucket}",
            f"bucket_file_offset={self.bucket_file_offset}",
        ]
        return f"<EvictionIterator [{', '.join(out)}]>"



"""
SEP: 0007
Title: URI Scheme to facilitate delegated signing
Author: Nikhil Saraf (Lightyear.io / SDF)
Status: Active
Created: 2018-05-07
Updated: 2020-03-03
Version: 2.0.0
"""

import abc
import base64
from decimal import Decimal
from typing import Dict, List, Optional, Tuple, Union
from urllib import parse

from ..asset import Asset
from ..fee_bump_transaction_envelope import FeeBumpTransactionEnvelope
from ..keypair import Keypair
from ..memo import HashMemo, IdMemo, Memo, NoneMemo, ReturnHashMemo, TextMemo
from ..transaction_envelope import TransactionEnvelope

__all__ = ["PayStellarUri", "TransactionStellarUri", "Replacement"]

STELLAR_SCHEME: str = "web+stellar"


class StellarUri(object, metaclass=abc.ABCMeta):
    def __init__(self, signature: Optional[str] = None):
        self.signature = signature

    @abc.abstractmethod
    def to_uri(self) -> str:
        raise NotImplementedError("The method has not been implemented.")

    @property
    def _signature_payload(self) -> bytes:
        data = self.to_uri()
        if self.signature:
            data = data[: data.find("&signature=")]
        # https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0007.md#request-signing
        sign_data = b"\0" * 35 + b"\4" + b"stellar.sep.7 - URI Scheme" + data.encode()
        return sign_data

    def sign(self, signer: Union[Keypair, str]) -> None:
        """Sign the URI.

        :param signer: The account used to sign this transaction, it should be the secret key of `URI_REQUEST_SIGNING_KEY`.
        """
        if isinstance(signer, str):
            signer = Keypair.from_secret(signer)
        sign_data = self._signature_payload
        signature = signer.sign(sign_data)
        self.signature = base64.b64encode(signature).decode()

    @staticmethod
    def _parse_uri_query(uri_query) -> Dict[str, str]:
        return dict(parse.parse_qsl(uri_query))

    @staticmethod
    def _parse_callback(callback: Optional[str]) -> Optional[str]:
        if callback is None:
            return None
        if not callback.startswith("url:"):
            raise ValueError("`callback` should start with `url:`.")
        return callback[4:]


class PayStellarUri(StellarUri):
    """A request for a payment to be signed.

    See `SEP-0007 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0007.md#operation-pay>`_

    :param destination: A valid account ID or payment address.
    :param amount: Amount that destination will receive.
    :param asset: Asset destination will receive.
    :param memo: A memo to attach to the transaction.
    :param callback: The uri to post the transaction to after signing.
    :param message: An message for displaying to the user.
    :param network_passphrase: The passphrase of the target network.
    :param origin_domain: A fully qualified domain name that specifies the originating domain of the URI request.
    :param signature: A base64 encode signature of the hash of the URI request.
    """

    def __init__(
        self,
        destination: str,
        amount: Union[str, Decimal, None] = None,
        asset: Optional[Asset] = None,
        memo: Optional[Memo] = None,
        callback: Optional[str] = None,
        message: Optional[str] = None,
        network_passphrase: Optional[str] = None,
        origin_domain: Optional[str] = None,
        signature: Optional[str] = None,
    ) -> None:
        super().__init__(signature)
        if message is not None and len(message) > 300:
            raise ValueError("Message must not exceed 300 characters.")

        self.asset_code = None
        self.asset_issuer = None
        self._asset = asset
        if asset:
            self.asset_code = asset.code
            self.asset_issuer = asset.issuer

        self.memo = None
        self.memo_type = None
        self._memo = memo
        self.memo_type, self.memo = self._encode_memo(memo)
        self.destination = destination
        if isinstance(amount, Decimal):
            amount = str(amount)
        self.amount = amount
        self.callback = callback
        self.msg = message
        self.network_passphrase = network_passphrase
        self.origin_domain = origin_domain

    @staticmethod
    def _encode_memo(memo) -> Union[Tuple[str, Union[str, int]], Tuple[None, None]]:
        if memo and not isinstance(memo, NoneMemo):
            if isinstance(memo, TextMemo):
                memo_text = memo.memo_text.decode()  # memo text cant decode?
                memo_type = "MEMO_TEXT"
                return memo_type, memo_text
            elif isinstance(memo, IdMemo):
                memo_id = memo.memo_id
                memo_type = "MEMO_ID"
                return memo_type, memo_id
            elif isinstance(memo, HashMemo):
                memo_hash = base64.b64encode(memo.memo_hash).decode()
                memo_type = "MEMO_HASH"
                return memo_type, memo_hash
            elif isinstance(memo, ReturnHashMemo):
                memo_return = base64.b64encode(memo.memo_return).decode()
                memo_type = "MEMO_RETURN"
                return memo_type, memo_return
            else:
                raise ValueError("Invalid memo.")
        return None, None

    @staticmethod
    def _decode_memo(
        memo_type: Optional[str], memo_value: Optional[str]
    ) -> Optional[Memo]:
        if memo_type is None:
            return None
        if memo_value is None:
            raise ValueError("`memo` is missing from uri.")
        if memo_type == "MEMO_TEXT":
            return TextMemo(memo_value)
        elif memo_type == "MEMO_ID":
            return IdMemo(int(memo_value))
        elif memo_type == "MEMO_HASH":
            value = base64.b64decode(memo_value.encode())
            return HashMemo(value)
        elif memo_type == "MEMO_RETURN":
            value = base64.b64decode(memo_value.encode())
            return ReturnHashMemo(value)
        else:
            raise ValueError("Invalid `memo_type`.")

    def to_uri(self) -> str:
        """Generate the request URI.

        :return: Stellar Pay URI.
        """
        query_params: Dict[str, Union[str, int, None]] = {}
        query_params["destination"] = self.destination
        if self.amount is not None:
            query_params["amount"] = self.amount
        if self.amount is not None:
            query_params["amount"] = self.amount
        if self._asset is not None and not self._asset.is_native():
            query_params["asset_code"] = self.asset_code
            query_params["asset_issuer"] = self.asset_issuer
        if self._memo is not None and not isinstance(self._memo, NoneMemo):
            assert self.memo is not None
            query_params["memo"] = self.memo
            query_params["memo_type"] = self.memo_type
        if self.callback is not None:
            query_params["callback"] = "url:" + self.callback
        if self.msg is not None:
            query_params["msg"] = self.msg
        if self.network_passphrase is not None:
            query_params["network_passphrase"] = self.network_passphrase
        if self.origin_domain is not None:
            query_params["origin_domain"] = self.origin_domain
        if self.signature is not None:
            query_params["signature"] = self.signature
        # https://github.com/python/typeshed/issues/4234
        return f"{STELLAR_SCHEME}:pay?{parse.urlencode(query_params, quote_via=parse.quote)}"  # type: ignore[type-var]

    @classmethod
    def from_uri(cls, uri: str) -> "PayStellarUri":
        """Parse Stellar Pay URI and generate :class:`PayStellarUri` object.

        :param uri: Stellar Pay URI.
        :return: :class:`PayStellarUri` object from uri.
        """
        parsed_uri = parse.urlparse(uri)
        if parsed_uri.scheme != STELLAR_SCHEME:
            raise ValueError(
                f"Stellar URI scheme should be `{STELLAR_SCHEME}`, but got `{parsed_uri.scheme}`."
            )
        if parsed_uri.path != "pay":
            raise ValueError(
                f"Stellar URI path should be `pay`, but got `{parsed_uri.path}`."
            )
        query = cls._parse_uri_query(parsed_uri.query)
        destination = query.get("destination")
        amount = query.get("amount")
        asset_code = query.get("asset_code")
        asset_issuer = query.get("asset_issuer")
        memo_value = query.get("memo")
        memo_type = query.get("memo_type")
        callback = cls._parse_callback(query.get("callback"))
        msg = query.get("msg")
        network_passphrase = query.get("network_passphrase")
        origin_domain = query.get("origin_domain")
        signature = query.get("signature")
        asset = None
        if asset_code is not None:
            asset = Asset(asset_code, asset_issuer)
        memo = cls._decode_memo(memo_type=memo_type, memo_value=memo_value)

        if destination is None:
            raise ValueError("`destination` is missing from uri.")

        return cls(
            destination=destination,
            amount=amount,
            asset=asset,
            memo=memo,
            callback=callback,
            message=msg,
            network_passphrase=network_passphrase,
            origin_domain=origin_domain,
            signature=signature,
        )

    def __repr__(self):
        return (
            f"<PayStellarUri [destination={self.destination}, amount={self.amount}, "
            f"asset_code={self.asset_code}, asset_issuer={self.asset_issuer}, "
            f"memo={self.memo}, memo_type={self.memo_type}, callback={self.callback}, "
            f"msg={self.msg}, network_passphrase={self.network_passphrase}, "
            f"origin_domain={self.origin_domain}, signature={self.signature}]>"
        )

    def __hash__(self):
        return hash(
            (
                self.destination,
                self.amount,
                self.asset_code,
                self.asset_issuer,
                self.memo,
                self.memo_type,
                self.callback,
                self.msg,
                self.network_passphrase,
                self.origin_domain,
                self.signature,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.destination == other.destination
            and self.amount == other.amount
            and self.asset_code == other.asset_code
            and self.asset_issuer == other.asset_issuer
            and self.memo == other.memo
            and self.memo_type == other.memo_type
            and self.callback == other.callback
            and self.msg == other.msg
            and self.network_passphrase == other.network_passphrase
            and self.origin_domain == other.origin_domain
            and self.signature == other.signature
        )


class Replacement:
    """Used to represent a single replacement.

    An example::

        r1 = Replacement("sourceAccount", "X", "account on which to create the trustline")
        r2 = Replacement("seqNum", "Y", "sequence for sourceAccount")
        replacements = [r1, r2]

    See `SEP-0007 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0007.md#operation-tx>`__

    :param txrep_tx_field_name: Txrep tx field name.
    :param reference_identifier: Reference identifier.
    :param hint: A brief and clear explanation of the context for the `reference_identifier`.

    """

    def __init__(
        self, txrep_tx_field_name: str, reference_identifier: str, hint: str
    ) -> None:
        self.txrep_tx_field_name = txrep_tx_field_name
        self.reference_identifier = reference_identifier
        self.hint = hint

    def __repr__(self):
        return (
            f"<Replacement [txrep_tx_field_name={self.txrep_tx_field_name}, "
            f"reference_identifier={self.reference_identifier}, "
            f"hint={self.hint}]>"
        )

    def __hash__(self):
        return hash((self.txrep_tx_field_name, self.reference_identifier, self.hint))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.txrep_tx_field_name == other.txrep_tx_field_name
            and self.reference_identifier == other.reference_identifier
            and self.hint == other.hint
        )


class TransactionStellarUri(StellarUri):
    """A request for a transaction to be signed.

    See `SEP-0007 <https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0007.md#operation-tx>`__

    :param transaction_envelope: Transaction waiting to be signed.
    :param replace: A value that identifies the fields to be replaced in the xdr using the Txrep (SEP-0011) representation.
    :param callback: The uri to post the transaction to after signing.
    :param pubkey: Specify which public key you want the URI handler to sign for.
    :param message: An message for displaying to the user.
    :param network_passphrase: The passphrase of the target network.
    :param origin_domain: A fully qualified domain name that specifies the originating domain of the URI request.
    :param signature: A base64 encode signature of the hash of the URI request.
    """

    def __init__(
        self,
        transaction_envelope: Union[TransactionEnvelope, FeeBumpTransactionEnvelope],
        replace: Optional[List[Replacement]] = None,
        callback: Optional[str] = None,
        pubkey: Optional[str] = None,
        message: Optional[str] = None,
        network_passphrase: Optional[str] = None,
        origin_domain: Optional[str] = None,
        signature: Optional[str] = None,
    ) -> None:
        super().__init__(signature)
        if message is not None and len(message) > 300:
            raise ValueError("Message must not exceed 300 characters.")

        self.transaction_envelope = transaction_envelope
        self.replace = replace
        self.callback = callback
        self.pubkey = pubkey
        self.msg = message
        self.network_passphrase = network_passphrase
        self.origin_domain = origin_domain

    @property
    def _replace(self) -> Optional[str]:
        if not self.replace:
            return None
        replaces = []
        hits = dict()
        for i in self.replace:
            hits[i.reference_identifier] = i.hint
            replaces.append(i.txrep_tx_field_name + ":" + i.reference_identifier)
        return (
            ",".join(replaces) + ";" + ",".join([k + ":" + v for k, v in hits.items()])
        )

    def to_uri(self) -> str:
        """Generate the request URI.

        :return: Stellar Transaction URI.
        """
        query_params = dict()
        query_params["xdr"] = self.transaction_envelope.to_xdr()
        if self.callback is not None:
            query_params["callback"] = "url:" + self.callback
        replace = self._replace
        if replace is not None:
            query_params["replace"] = replace
        if self.pubkey is not None:
            query_params["pubkey"] = self.pubkey
        if self.msg is not None:
            query_params["msg"] = self.msg
        if self.network_passphrase is not None:
            query_params["network_passphrase"] = self.network_passphrase
        if self.origin_domain is not None:
            query_params["origin_domain"] = self.origin_domain
        if self.signature is not None:
            query_params["signature"] = self.signature
        return f"{STELLAR_SCHEME}:tx?{parse.urlencode(query_params, quote_via=parse.quote)}"  # type: ignore[type-var]

    @classmethod
    def from_uri(
        cls, uri: str, network_passphrase: Optional[str]
    ) -> "TransactionStellarUri":
        """Parse Stellar Transaction URI and generate :class:`TransactionStellarUri` object.

        :param uri: Stellar Transaction URI.
        :param network_passphrase: The network to connect to for verifying and retrieving xdr,
            If it is set to `None`, the `network_passphrase` in the uri will not be verified.
        :return: :class:`TransactionStellarUri` object from uri.
        """
        parsed_uri = parse.urlparse(uri)
        if parsed_uri.scheme != STELLAR_SCHEME:
            raise ValueError(
                f"Stellar URI scheme should be `{STELLAR_SCHEME}`, but got `{parsed_uri.scheme}`."
            )
        if parsed_uri.path != "tx":
            raise ValueError(
                f"Stellar URI path should be `tx`, but got `{parsed_uri.path}`."
            )
        query = cls._parse_uri_query(parsed_uri.query)
        uri_network_passphrase = query.get("network_passphrase")
        if network_passphrase is None and uri_network_passphrase is None:
            raise ValueError("`network_passphrase` is required.")

        if (
            uri_network_passphrase is not None
            and network_passphrase is not None
            and network_passphrase != uri_network_passphrase
        ):
            raise ValueError(
                "The `network_passphrase` in the function parameter does not "
                "match the `network_passphrase` in the uri."
            )
        network_passphrase = network_passphrase or uri_network_passphrase
        assert network_passphrase is not None

        xdr = query.get("xdr")
        callback = cls._parse_callback(query.get("callback"))
        pubkey = query.get("pubkey")
        msg = query.get("msg")
        origin_domain = query.get("origin_domain")
        signature = query.get("signature")
        if xdr is None:
            raise ValueError("`xdr` is missing from uri.")
        if FeeBumpTransactionEnvelope.is_fee_bump_transaction_envelope(xdr):
            tx = FeeBumpTransactionEnvelope.from_xdr(xdr, network_passphrase)
        else:
            tx = TransactionEnvelope.from_xdr(xdr, network_passphrase)  # type: ignore[assignment]
        raw_replacements = query.get("replace")
        replacements = []
        if raw_replacements is not None:
            descriptions_map = {}
            identifiers, descriptions = raw_replacements.split(";")
            for description in descriptions.split(","):
                k, v = description.split(":")
                descriptions_map[k] = v
            for identifier in identifiers.split(","):
                k, v = identifier.split(":")
                hint = descriptions_map.get(v)
                if hint is None:
                    raise ValueError("Invalid `replace`.")
                replacement = Replacement(k, v, hint)
                replacements.append(replacement)
        return cls(
            transaction_envelope=tx,
            replace=replacements,
            callback=callback,
            pubkey=pubkey,
            message=msg,
            network_passphrase=network_passphrase,
            origin_domain=origin_domain,
            signature=signature,
        )

    def __repr__(self):
        return (
            f"<TransactionStellarUri [xdr={self.transaction_envelope.to_xdr()}, replace={self.replace}, "
            f"callback={self.callback}, pubkey={self.pubkey}, "
            f"msg={self.msg}, network_passphrase={self.network_passphrase}, "
            f"origin_domain={self.origin_domain}, signature={self.signature}]>"
        )

    def __hash__(self):
        return hash(
            (
                self.transaction_envelope,
                self.replace,
                self.callback,
                self.pubkey,
                self.msg,
                self.network_passphrase,
                self.origin_domain,
                self.signature,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.transaction_envelope == other.transaction_envelope
            and self.replace == other.replace
            and self.callback == other.callback
            and self.pubkey == other.pubkey
            and self.msg == other.msg
            and self.network_passphrase == other.network_passphrase
            and self.origin_domain == other.origin_domain
            and self.signature == other.signature
        )



"""This file contains constants, functions and classes used internally by this SDK.
They may change at any time, so please do not use them directly.
"""

import hashlib
import re
from decimal import ROUND_FLOOR, Context, Decimal, Inexact
from typing import TYPE_CHECKING, Dict, Optional, Sequence, Union
from urllib.parse import urljoin, urlsplit, urlunsplit

from .exceptions import Ed25519PublicKeyInvalidError, NoApproximationError
from .strkey import StrKey

if TYPE_CHECKING:
    from .asset import Asset

MUXED_ACCOUNT_STARTING_LETTER: str = "M"
ED25519_PUBLIC_KEY_STARTING_LETTER: str = "G"
_LOWER_LIMIT = "0"
_UPPER_LIMIT = "922337203685.4775807"
_EXPONENT = 7
_ONE = Decimal(10**7)


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def best_rational_approximation(x) -> Dict[str, int]:
    x = Decimal(x)
    int32_max = Decimal(2147483647)
    fractions = [[Decimal(0), Decimal(1)], [Decimal(1), Decimal(0)]]
    i = 2
    while True:
        if x > int32_max:
            break
        a = x.to_integral_exact(rounding=ROUND_FLOOR)
        f = x - a
        h = a * fractions[i - 1][0] + fractions[i - 2][0]
        k = a * fractions[i - 1][1] + fractions[i - 2][1]
        if h > int32_max or k > int32_max:
            break
        fractions.append([h, k])
        if f.is_zero():
            break
        x = 1 / f
        i = i + 1
    n = fractions[len(fractions) - 1][0]
    d = fractions[len(fractions) - 1][1]
    if n.is_zero() or d.is_zero():
        raise NoApproximationError("Couldn't find approximation.")
    return {"n": int(n), "d": int(d)}


def hex_to_bytes(hex_string: Union[str, bytes]) -> bytes:
    if isinstance(hex_string, str):
        return bytes.fromhex(hex_string)
    return hex_string


def convert_assets_to_horizon_param(assets: Sequence["Asset"]) -> str:
    assets_string = []
    for asset in assets:
        if asset.is_native():
            assets_string.append(asset.type)
        else:
            assets_string.append(f"{asset.code}:{asset.issuer}")
    return ",".join(assets_string)


def urljoin_with_query(base: str, path: Optional[str]) -> str:
    split_url = urlsplit(base)
    query = split_url.query
    real_path = split_url.path
    if path:
        if not real_path.endswith("/"):
            real_path += "/"
        real_path = urljoin(real_path, path)
    url = urlunsplit(
        (split_url.scheme, split_url.netloc, real_path, query, split_url.fragment)
    )
    return url


def is_valid_hash(data: str) -> bool:
    if not data:
        return False
    asset_code_re = re.compile(r"^[a-zA-Z0-9]{64}$")
    return bool(asset_code_re.match(data))


def raise_if_not_valid_ed25519_public_key(value: str, argument_name: str) -> None:
    try:
        StrKey.decode_ed25519_public_key(value)
    except Exception as e:
        raise Ed25519PublicKeyInvalidError(
            f'Value of argument "{argument_name}" is not a valid ed25519 public key: {value}'
        ) from e


def raise_if_not_valid_amount(value: str, argument_name: str) -> None:
    amount = Decimal(value)
    exponent = amount.as_tuple().exponent
    if not isinstance(exponent, int) or abs(exponent) > _EXPONENT:
        raise ValueError(
            f'Value of argument "{argument_name}" must have at most 7 digits after the decimal: {amount}'
        )
    if amount < Decimal(_LOWER_LIMIT) or amount > Decimal(_UPPER_LIMIT):
        raise ValueError(
            f'Value of argument "{argument_name}" must represent a positive number '
            f"and the max valid value is {_UPPER_LIMIT}: {amount}"
        )


def raise_if_not_valid_hash(value: str, argument_name: str) -> None:
    if not is_valid_hash(value):
        raise ValueError(
            f'Value of argument "{argument_name}" is not a valid hash: {value}'
        )


def raise_if_not_valid_balance_id(value: str, argument_name: str) -> None:
    if len(value) != 72 or value[:8] != "00000000" or not is_valid_hash(value[8:]):
        raise ValueError(
            f'Value of argument "{argument_name}" is not a valid balance id: {value}'
        )


def to_xdr_amount(value: Union[str, Decimal]) -> int:
    """Converts an amount to the appropriate value to send over the network
    as a part of an XDR object.

    Each asset amount is encoded as a signed 64-bit integer in the XDR
    structures. An asset amount unit (that which is seen by end users) is
    scaled down by a factor of ten million (10,000,000) to arrive at the
    native 64-bit integer representation. For example, the integer amount
    value 25,123,456 equals 2.5123456 units of the asset. This scaling
    allows for seven decimal places of precision in human-friendly amount
    units.

    This static method correctly multiplies the value by the scaling factor
    in order to come to the integer value used in XDR structures.

    See `Stellar's documentation on Asset Precision
    <https://developers.stellar.org/docs/issuing-assets/anatomy-of-an-asset/#amount-precision>`_
    for more information.

    :param value: The amount to convert to an integer for XDR
        serialization.

    """
    # throw exception if value * ONE has decimal places (it can't be represented as int64)
    try:
        amount = int(
            (Decimal(value) * _ONE).to_integral_exact(context=Context(traps=[Inexact]))
        )
    except Inexact:
        raise ValueError(
            f"Value of '{value}' must have at most 7 digits after the decimal."
        )

    if amount < 0 or amount > 9223372036854775807:
        raise ValueError(
            f"Value of '{value}' must represent a positive number "
            "and the max valid value is 922337203685.4775807."
        )

    return amount


def from_xdr_amount(value: int) -> str:
    """Converts a str amount from an XDR amount object

    :param value: The amount to convert to a string from an XDR int64
        amount.

    """
    return str(Decimal(value) / _ONE)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["AccountMergeResultCode"]


class AccountMergeResultCode(IntEnum):
    """
    XDR Source Code::

        enum AccountMergeResultCode
        {
            // codes considered as "success" for the operation
            ACCOUNT_MERGE_SUCCESS = 0,
            // codes considered as "failure" for the operation
            ACCOUNT_MERGE_MALFORMED = -1,       // can't merge onto itself
            ACCOUNT_MERGE_NO_ACCOUNT = -2,      // destination does not exist
            ACCOUNT_MERGE_IMMUTABLE_SET = -3,   // source account has AUTH_IMMUTABLE set
            ACCOUNT_MERGE_HAS_SUB_ENTRIES = -4, // account has trust lines/offers
            ACCOUNT_MERGE_SEQNUM_TOO_FAR = -5,  // sequence number is over max allowed
            ACCOUNT_MERGE_DEST_FULL = -6,       // can't add source balance to
                                                // destination balance
            ACCOUNT_MERGE_IS_SPONSOR = -7       // can't merge account that is a sponsor
        };
    """

    ACCOUNT_MERGE_SUCCESS = 0
    ACCOUNT_MERGE_MALFORMED = -1
    ACCOUNT_MERGE_NO_ACCOUNT = -2
    ACCOUNT_MERGE_IMMUTABLE_SET = -3
    ACCOUNT_MERGE_HAS_SUB_ENTRIES = -4
    ACCOUNT_MERGE_SEQNUM_TOO_FAR = -5
    ACCOUNT_MERGE_DEST_FULL = -6
    ACCOUNT_MERGE_IS_SPONSOR = -7

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AccountMergeResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AccountMergeResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AccountMergeResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["AccountFlags"]


class AccountFlags(IntEnum):
    """
    XDR Source Code::

        enum AccountFlags
        { // masks for each flag

            // Flags set on issuer accounts
            // TrustLines are created with authorized set to "false" requiring
            // the issuer to set it for each TrustLine
            AUTH_REQUIRED_FLAG = 0x1,
            // If set, the authorized flag in TrustLines can be cleared
            // otherwise, authorization cannot be revoked
            AUTH_REVOCABLE_FLAG = 0x2,
            // Once set, causes all AUTH_* flags to be read-only
            AUTH_IMMUTABLE_FLAG = 0x4,
            // Trustlines are created with clawback enabled set to "true",
            // and claimable balances created from those trustlines are created
            // with clawback enabled set to "true"
            AUTH_CLAWBACK_ENABLED_FLAG = 0x8
        };
    """

    AUTH_REQUIRED_FLAG = 1
    AUTH_REVOCABLE_FLAG = 2
    AUTH_IMMUTABLE_FLAG = 4
    AUTH_CLAWBACK_ENABLED_FLAG = 8

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AccountFlags:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AccountFlags:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AccountFlags:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..muxed_account import MuxedAccount
from ..price import Price
from ..utils import raise_if_not_valid_amount
from .operation import Operation

__all__ = ["ManageSellOffer"]


class ManageSellOffer(Operation):
    """The :class:`ManageSellOffer` object, which represents a ManageSellOffer
    operation on Stellar's network.

    Creates, updates, or deletes an sell offer.

    If you want to create a new offer set `offer_id` to ``0``.

    If you want to update an existing offer set `offer_id` to existing offer ID.

    If you want to delete an existing offer set `offer_id` to existing offer ID
    and set `amount` to ``0``.

    Threshold: Medium

    See `Manage Sell Offer <https://developers.stellar.org/docs/start/list-of-operations/#manage-sell-offer>`_ for more information.

    :param selling: What you're selling.
    :param buying: What you're buying.
    :param amount: The total amount you're selling. If ``0``, deletes the offer.
    :param price: Price of 1 unit of `selling` in terms of `buying`.
    :param offer_id: If ``0``, will create a new offer (default). Otherwise,
        edits an existing offer.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.MANAGE_SELL_OFFER
    )

    def __init__(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        offer_id: int = 0,
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        if not isinstance(price, Price):
            price = Price.from_raw_price(price)
        self.selling: Asset = selling
        self.buying: Asset = buying
        self.amount: str = str(amount)
        self.price: Price = price
        self.offer_id: int = offer_id
        raise_if_not_valid_amount(self.amount, "amount")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        selling = self.selling.to_xdr_object()
        buying = self.buying.to_xdr_object()
        price = self.price.to_xdr_object()
        amount = stellar_xdr.Int64(Operation.to_xdr_amount(self.amount))
        manage_sell_offer_op = stellar_xdr.ManageSellOfferOp(
            selling, buying, amount, price, stellar_xdr.Int64(self.offer_id)
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE, manage_sell_offer_op=manage_sell_offer_op
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "ManageSellOffer":
        """Creates a :class:`ManageSellOffer` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.manage_sell_offer_op is not None
        selling = Asset.from_xdr_object(xdr_object.body.manage_sell_offer_op.selling)
        buying = Asset.from_xdr_object(xdr_object.body.manage_sell_offer_op.buying)
        amount = Operation.from_xdr_amount(
            xdr_object.body.manage_sell_offer_op.amount.int64
        )
        price = Price.from_xdr_object(xdr_object.body.manage_sell_offer_op.price)
        offer_id = xdr_object.body.manage_sell_offer_op.offer_id.int64

        op = cls(
            source=source,
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
        )
        return op

    def __repr__(self):
        return (
            f"<ManageSellOffer [selling={self.selling}, buying={self.buying}, "
            f"amount={self.amount}, price={self.price}, offer_id={self.offer_id}, source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .contract_executable_type import ContractExecutableType
from .hash import Hash

__all__ = ["ContractExecutable"]


class ContractExecutable:
    """
    XDR Source Code::

        union ContractExecutable switch (ContractExecutableType type)
        {
        case CONTRACT_EXECUTABLE_WASM:
            Hash wasm_hash;
        case CONTRACT_EXECUTABLE_STELLAR_ASSET:
            void;
        };
    """

    def __init__(
        self,
        type: ContractExecutableType,
        wasm_hash: Hash = None,
    ) -> None:
        self.type = type
        self.wasm_hash = wasm_hash

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == ContractExecutableType.CONTRACT_EXECUTABLE_WASM:
            if self.wasm_hash is None:
                raise ValueError("wasm_hash should not be None.")
            self.wasm_hash.pack(packer)
            return
        if self.type == ContractExecutableType.CONTRACT_EXECUTABLE_STELLAR_ASSET:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractExecutable:
        type = ContractExecutableType.unpack(unpacker)
        if type == ContractExecutableType.CONTRACT_EXECUTABLE_WASM:
            wasm_hash = Hash.unpack(unpacker)
            return cls(type=type, wasm_hash=wasm_hash)
        if type == ContractExecutableType.CONTRACT_EXECUTABLE_STELLAR_ASSET:
            return cls(type=type)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractExecutable:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractExecutable:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.wasm_hash,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.type == other.type and self.wasm_hash == other.wasm_hash

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"wasm_hash={self.wasm_hash}")
            if self.wasm_hash is not None
            else None
        )
        return f"<ContractExecutable [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .liabilities import Liabilities
from .trust_line_entry_v1_ext import TrustLineEntryV1Ext

__all__ = ["TrustLineEntryV1"]


class TrustLineEntryV1:
    """
    XDR Source Code::

        struct
                {
                    Liabilities liabilities;

                    union switch (int v)
                    {
                    case 0:
                        void;
                    case 2:
                        TrustLineEntryExtensionV2 v2;
                    }
                    ext;
                }
    """

    def __init__(
        self,
        liabilities: Liabilities,
        ext: TrustLineEntryV1Ext,
    ) -> None:
        self.liabilities = liabilities
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.liabilities.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TrustLineEntryV1:
        liabilities = Liabilities.unpack(unpacker)
        ext = TrustLineEntryV1Ext.unpack(unpacker)
        return cls(
            liabilities=liabilities,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TrustLineEntryV1:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TrustLineEntryV1:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.liabilities,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.liabilities == other.liabilities and self.ext == other.ext

    def __repr__(self):
        out = [
            f"liabilities={self.liabilities}",
            f"ext={self.ext}",
        ]
        return f"<TrustLineEntryV1 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .peer_stat_list import PeerStatList
from .uint32 import Uint32

__all__ = ["TopologyResponseBodyV0"]


class TopologyResponseBodyV0:
    """
    XDR Source Code::

        struct TopologyResponseBodyV0
        {
            PeerStatList inboundPeers;
            PeerStatList outboundPeers;

            uint32 totalInboundPeerCount;
            uint32 totalOutboundPeerCount;
        };
    """

    def __init__(
        self,
        inbound_peers: PeerStatList,
        outbound_peers: PeerStatList,
        total_inbound_peer_count: Uint32,
        total_outbound_peer_count: Uint32,
    ) -> None:
        self.inbound_peers = inbound_peers
        self.outbound_peers = outbound_peers
        self.total_inbound_peer_count = total_inbound_peer_count
        self.total_outbound_peer_count = total_outbound_peer_count

    def pack(self, packer: Packer) -> None:
        self.inbound_peers.pack(packer)
        self.outbound_peers.pack(packer)
        self.total_inbound_peer_count.pack(packer)
        self.total_outbound_peer_count.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TopologyResponseBodyV0:
        inbound_peers = PeerStatList.unpack(unpacker)
        outbound_peers = PeerStatList.unpack(unpacker)
        total_inbound_peer_count = Uint32.unpack(unpacker)
        total_outbound_peer_count = Uint32.unpack(unpacker)
        return cls(
            inbound_peers=inbound_peers,
            outbound_peers=outbound_peers,
            total_inbound_peer_count=total_inbound_peer_count,
            total_outbound_peer_count=total_outbound_peer_count,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TopologyResponseBodyV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TopologyResponseBodyV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.inbound_peers,
                self.outbound_peers,
                self.total_inbound_peer_count,
                self.total_outbound_peer_count,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.inbound_peers == other.inbound_peers
            and self.outbound_peers == other.outbound_peers
            and self.total_inbound_peer_count == other.total_inbound_peer_count
            and self.total_outbound_peer_count == other.total_outbound_peer_count
        )

    def __repr__(self):
        out = [
            f"inbound_peers={self.inbound_peers}",
            f"outbound_peers={self.outbound_peers}",
            f"total_inbound_peer_count={self.total_inbound_peer_count}",
            f"total_outbound_peer_count={self.total_outbound_peer_count}",
        ]
        return f"<TopologyResponseBodyV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .node_id import NodeID
from .uint32 import Uint32

__all__ = ["TimeSlicedSurveyStopCollectingMessage"]


class TimeSlicedSurveyStopCollectingMessage:
    """
    XDR Source Code::

        struct TimeSlicedSurveyStopCollectingMessage
        {
            NodeID surveyorID;
            uint32 nonce;
            uint32 ledgerNum;
        };
    """

    def __init__(
        self,
        surveyor_id: NodeID,
        nonce: Uint32,
        ledger_num: Uint32,
    ) -> None:
        self.surveyor_id = surveyor_id
        self.nonce = nonce
        self.ledger_num = ledger_num

    def pack(self, packer: Packer) -> None:
        self.surveyor_id.pack(packer)
        self.nonce.pack(packer)
        self.ledger_num.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TimeSlicedSurveyStopCollectingMessage:
        surveyor_id = NodeID.unpack(unpacker)
        nonce = Uint32.unpack(unpacker)
        ledger_num = Uint32.unpack(unpacker)
        return cls(
            surveyor_id=surveyor_id,
            nonce=nonce,
            ledger_num=ledger_num,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TimeSlicedSurveyStopCollectingMessage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TimeSlicedSurveyStopCollectingMessage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.surveyor_id,
                self.nonce,
                self.ledger_num,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.surveyor_id == other.surveyor_id
            and self.nonce == other.nonce
            and self.ledger_num == other.ledger_num
        )

    def __repr__(self):
        out = [
            f"surveyor_id={self.surveyor_id}",
            f"nonce={self.nonce}",
            f"ledger_num={self.ledger_num}",
        ]
        return f"<TimeSlicedSurveyStopCollectingMessage [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .claimable_balance_id import ClaimableBalanceID

__all__ = ["ClawbackClaimableBalanceOp"]


class ClawbackClaimableBalanceOp:
    """
    XDR Source Code::

        struct ClawbackClaimableBalanceOp
        {
            ClaimableBalanceID balanceID;
        };
    """

    def __init__(
        self,
        balance_id: ClaimableBalanceID,
    ) -> None:
        self.balance_id = balance_id

    def pack(self, packer: Packer) -> None:
        self.balance_id.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClawbackClaimableBalanceOp:
        balance_id = ClaimableBalanceID.unpack(unpacker)
        return cls(
            balance_id=balance_id,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClawbackClaimableBalanceOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClawbackClaimableBalanceOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.balance_id,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.balance_id == other.balance_id

    def __repr__(self):
        out = [
            f"balance_id={self.balance_id}",
        ]
        return f"<ClawbackClaimableBalanceOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .crypto_key_type import CryptoKeyType
from .muxed_account_med25519 import MuxedAccountMed25519
from .uint256 import Uint256

__all__ = ["MuxedAccount"]


class MuxedAccount:
    """
    XDR Source Code::

        union MuxedAccount switch (CryptoKeyType type)
        {
        case KEY_TYPE_ED25519:
            uint256 ed25519;
        case KEY_TYPE_MUXED_ED25519:
            struct
            {
                uint64 id;
                uint256 ed25519;
            } med25519;
        };
    """

    def __init__(
        self,
        type: CryptoKeyType,
        ed25519: Uint256 = None,
        med25519: MuxedAccountMed25519 = None,
    ) -> None:
        self.type = type
        self.ed25519 = ed25519
        self.med25519 = med25519

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == CryptoKeyType.KEY_TYPE_ED25519:
            if self.ed25519 is None:
                raise ValueError("ed25519 should not be None.")
            self.ed25519.pack(packer)
            return
        if self.type == CryptoKeyType.KEY_TYPE_MUXED_ED25519:
            if self.med25519 is None:
                raise ValueError("med25519 should not be None.")
            self.med25519.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> MuxedAccount:
        type = CryptoKeyType.unpack(unpacker)
        if type == CryptoKeyType.KEY_TYPE_ED25519:
            ed25519 = Uint256.unpack(unpacker)
            return cls(type=type, ed25519=ed25519)
        if type == CryptoKeyType.KEY_TYPE_MUXED_ED25519:
            med25519 = MuxedAccountMed25519.unpack(unpacker)
            return cls(type=type, med25519=med25519)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> MuxedAccount:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> MuxedAccount:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.ed25519,
                self.med25519,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.ed25519 == other.ed25519
            and self.med25519 == other.med25519
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"ed25519={self.ed25519}") if self.ed25519 is not None else None
        out.append(f"med25519={self.med25519}") if self.med25519 is not None else None
        return f"<MuxedAccount [{', '.join(out)}]>"
